from .dtc import DTC  # noqa
from .gradient_oinfo import GradientOinfo  # noqa
from .infotopo import InfoTopo  # noqa
from .info_tot import InfoTot  # noqa
from .oinfo import Oinfo  # noqa
from .phiid_syn import SynergyphiID  # noqa
from .phiid_red import RedundancyphiID  # noqa
from .red_mmi import RedundancyMMI  # noqa
from .rsi import RSI  # noqa
from .sinfo import Sinfo  # noqa
from .syn_mmi import SynergyMMI  # noqa
from .tc import TC  # noqa



import jax.numpy as jnp
import numpy as np
import itertools
from math import comb as ccomb


def _combinations(n, k, order, target):
    for c in itertools.combinations(range(n), k):
        # convert to list
        c = list(c) + target

        # deal with order
        if order:
            c = len(c)

        yield c


def combinations(
    n,
    minsize,
    maxsize=None,
    astype="iterator",
    order=False,
    fill_value=-1,
    target=None,
):
    """Get combinations.

    Parameters
    ----------
    n : int
        Represents the total number of elements in the set
    minsize : int
        Minimum size of the multiplets
    maxsize : int | None
        Maximum size of the multiplets. If None, minsize is used.
    astype : {'jax', 'numpy', 'iterator'}
        Specify the output type. Use either 'jax' get the data as a jax
        array [default], 'numpy' for NumPy array or 'iterator'.
    order : bool, optional
        If True, return the order of each multiplet. Default is False.
    fill_value : int | -1
        Value to use to fill missing values of multiplets.
    target : list | None
        List of indices to use as targets.

    Returns
    -------
    combs : jnp.array
        An array of shape (n_combinations, k) representing all possible
        combinations of k elements.
    """
    # ________________________________ ITERATOR _______________________________
    if not isinstance(maxsize, int):
        maxsize = minsize
    target = [] if target is None else list(target)
    assert maxsize >= minsize
    iterators = []
    for msize in range(minsize, maxsize + 1):
        iterators.append(_combinations(n, msize, order, target))
    iterators = itertools.chain(*tuple(iterators))

    if astype == "iterator":
        return iterators

    # _________________________________ ARRAYS ________________________________
    if order:
        combs = np.asarray([c for c in iterators]).astype(int)
    else:
        # get the number of combinations
        n_mults = sum([ccomb(n, c) for c in range(minsize, maxsize + 1)])

        # prepare output
        combs = np.full(
            (n_mults, maxsize + len(target)), fill_value, dtype=int
        )

        # fill the array
        for n_c, c in enumerate(iterators):
            combs[n_c, 0 : len(c)] = c

    # jax conversion (f required)
    if astype == "jax":
        combs = jnp.asarray(combs)

    return combs


if __name__ == "__main__":
    combs = combinations(
        10, minsize=1, maxsize=3, astype="jax", order=False, target=[11, 12]
    )
    print(combs)



from functools import partial

import numpy as np
import jax
import jax.numpy as jnp

from hoi.core.combinatory import combinations
from hoi.core.entropies import get_entropy, prepare_for_it
from hoi.utils.progressbar import get_pbar
from hoi.utils.logging import logger, set_log_level


@partial(jax.jit, static_argnums=(2,))
def ent_at_index(x, idx, entropy=None):
    """Compute entropy for a specific multiplet.

    This function has to be wrapped with the entropy function.
    """
    return x, entropy(x[:, idx, :])


class HOIEstimator(object):
    def __init__(self, x, y=None, multiplets=None, verbose=None):
        set_log_level(verbose)

        # check types of default properties
        _prop_lst = ["null", "redundancy", "synergy", "info"]
        assert isinstance(self.__name__, str)
        assert isinstance(self._encoding, bool)
        assert self._positive in _prop_lst
        assert self._negative in _prop_lst
        assert isinstance(self._symmetric, bool)

        # prepare the data
        self._x = self._preprocessing(x, y=y, multiplets=multiplets)

    def __iter__(self):
        """Iteration over orders."""
        for msize in range(self.minsize, self.maxsize + 1):
            yield msize

    ###########################################################################
    ###########################################################################
    #                                 I/O
    ###########################################################################
    ###########################################################################

    def _preprocessing(self, x, y=None, multiplets=None):
        """Check input x shape."""

        logger.debug("    Preprocessing of the data")

        # ---------------------------- X RESHAPING ----------------------------
        # force x to be 3d
        assert x.ndim >= 2
        if x.ndim == 2:
            x = x[..., np.newaxis]
        self._n_features_x = x.shape[1]
        self._n_features_y = 0

        # ---------------------------- Y RESHAPING ----------------------------
        # additional variable along feature dimension
        self._has_target = isinstance(y, (list, np.ndarray, tuple))
        if self._has_target:
            x = self._merge_xy(x, y=y)

        # ----------------------------- TRANSPOSE -----------------------------
        x = x.transpose(2, 1, 0)
        self.n_variables, self.n_features, self.n_samples = x.shape

        # ----------------------------- MULTIPLETS ----------------------------
        # compute only selected multiplets
        self._custom_mults = None
        if isinstance(multiplets, (list, np.ndarray)):
            self._custom_mults = multiplets

        return x

    def _merge_xy(self, x, y=None):
        """Merge x and y variables."""
        assert x.ndim == 3
        n_samples, n_features, n_variables = x.shape

        if y is None:
            return y

        y = np.asarray(y)

        # trial checking
        if y.shape[0] != n_samples:
            raise IOError(
                f"The numer of sample of the variable y ({y.shape[0]}) should "
                f"match the number of sample of the variable x ({n_samples})"
            )

        # 1d / 2d / 3d merging
        if y.ndim == 1:
            y = y[:, np.newaxis]
        if y.ndim == 2:
            y = np.tile(y[:, :, np.newaxis], n_variables)
        if y.shape[2] != n_variables:
            raise IOError(
                f"The numer of variables of the variable y ({y.shape[2]}) "
                f"should match the number of variables of the variable x"
                f" ({n_variables})"
            )
        self._n_features_x = x.shape[1]
        self._n_features_y = y.shape[1]

        return np.concatenate((x, y), axis=1)

    def _split_xy(self, xy):
        """Split back the xy variables into x and y."""
        return (
            xy[:, 0 : self._n_features_x, :],
            xy[:, -self._n_features_y : :, :],
        )

    def _check_minmax(self, minsize, maxsize):
        """Define min / max size of the multiplets."""

        # check minsize / maxsize
        if not isinstance(minsize, int):
            minsize = 1
        if not isinstance(maxsize, int):
            maxsize = self.n_features
        assert isinstance(maxsize, int)
        assert isinstance(minsize, int)
        assert maxsize >= minsize
        maxsize = max(1, min(maxsize, self.n_features))
        minsize = max(1, minsize)

        self.minsize, self.maxsize = minsize, maxsize

        return minsize, maxsize

    ###########################################################################
    ###########################################################################
    #                         INFORMATION THEORY
    ###########################################################################
    ###########################################################################
    def compute_entropies(
        self, method="gc", minsize=1, maxsize=None, samples=None, **kwargs
    ):
        """Compute entropies for all multiplets.

        Parameters
        ----------
        method : {'gc', 'binning', 'knn', 'kernel}
            Name of the method to compute entropy. Use either :

                * 'gc': gaussian copula entropy [default]. See
                  :func:`hoi.core.entropy_gc`
                * 'binning': binning-based estimator of entropy. Note that to
                  use this estimator, the data have be to discretized. See
                  :func:`hoi.core.entropy_bin`
                * 'knn': k-nearest neighbor estimator. See
                  :func:`hoi.core.entropy_knn`
                * 'kernel': kernel-based estimator of entropy
                  see :func:`hoi.core.entropy_kernel`

        samples : np.ndarray
            List of samples to use to compute HOI. If None, all samples are
            going to be used.
        minsize : int, optional
            Minimum size of the multiplets. Default is 1.
        maxsize : int, optional
            Maximum size of the multiplets. Default is None.
        kwargs : dict, optional
            Additional arguments to pass to the entropy function.

        Returns
        -------
        h_x : array_like
            Entropies of shape (n_mult, n_variables)
        h_idx : array_like
            Indices of the multiplets of shape (n_mult, maxsize)
        order : array_like
            Order of each multiplet of shape (n_mult,)
        """
        logger.info(f"Compute entropy with {method}")
        msg = "Entropy H(%i)"

        # ________________________________ I/O ________________________________
        # prepare the data for computing entropy
        x, kwargs = prepare_for_it(self._x, method, samples=samples, **kwargs)

        # get entropy function
        entropy = partial(
            ent_at_index,
            entropy=jax.vmap(get_entropy(method=method, **kwargs)),
        )

        # ______________________________ ENTROPY ______________________________
        # get all of the combinations
        h_idx, order = self.get_combinations(minsize, maxsize=maxsize)

        # get progress bar
        pbar = get_pbar(iterable=range(minsize, maxsize + 1), leave=False)

        # compute entropies
        offset = 0
        h_x = jnp.zeros((len(order), self.n_variables), dtype=jnp.float32)
        for msize in pbar:
            pbar.set_description(desc=msg % msize, refresh=False)

            # get the number of features when considering y
            n_feat_xy = msize + self._n_features_y

            # combinations of features
            _h_idx = h_idx[order == msize, 0:n_feat_xy]

            # compute all entropies
            _, _h_x = jax.lax.scan(entropy, x, _h_idx)

            # fill entropies
            n_combs = _h_idx.shape[0]
            h_x = h_x.at[offset : offset + n_combs, :].set(_h_x)

            offset += n_combs

        pbar.close()

        self._entropies = h_x

        return h_x, h_idx, order

    ###########################################################################
    ###########################################################################
    #                             COMPUTATIONS
    ###########################################################################
    ###########################################################################

    def get_combinations(self, minsize, maxsize=None, astype="jax"):
        """Get combinations of features.

        Parameters
        ----------
        minsize : int
            Minimum size of the multiplets
        maxsize : int | None
            Maximum size of the multiplets. If None, minsize is used.
        astype : {'jax', 'numpy', 'iterator'}
            Specify the output type. Use either 'jax' get the data as a jax
            array [default], 'numpy' for NumPy array or 'iterator'.

        Returns
        -------
        combinations : array_like
            Combinations of features.
        """
        logger.info("Get list of multiplets")
        # specify whether to include target or not
        n = self._n_features_x
        if (not self._has_target) or self._encoding:
            target = None
        else:
            target = (np.arange(self._n_features_y) + n).tolist()

        # custom list of multiplets don't require to full list of multiplets
        if self._custom_mults is not None:
            # for encoding or without target (for the other cases it is
            # automatic)
            if self._encoding or not self._has_target:
                target = []
            logger.info("    Selecting custom multiplets")
            _orders = [len(m) for m in self._custom_mults]
            # adding target for not encoding metrics
            mults_ = [list(c) + target for c in self._custom_mults]
            mults = jnp.full(
                (len(self._custom_mults), max(_orders) + len(target)), -1
            )
            for n_m, m in enumerate(mults_):
                mults = mults.at[n_m, 0 : len(m)].set(m)
            self._multiplets = mults
        else:
            # get the full list of multiplets
            self._multiplets = combinations(
                n,
                minsize,
                maxsize=maxsize,
                target=target,
                astype=astype,
                order=False,
                fill_value=-1,
            )

        return self._multiplets, self.order

    def fit(self):  # noqa
        raise NotImplementedError()

    ###########################################################################
    ###########################################################################
    #                             PROPERTIES
    ###########################################################################
    ###########################################################################

    @property
    def entropies(self):
        """Entropies of shape (n_mult,)"""
        return np.asarray(self._entropies)

    @property
    def multiplets(self):
        """Indices of the multiplets of shape (n_mult, maxsize).

        By convention, we used -1 to indicate that a feature has been ignored.
        """
        return np.asarray(self._multiplets)

    @property
    def order(self):
        """Order of each multiplet of shape (n_mult,)."""
        order = (self.multiplets >= 0).sum(1)
        if self._encoding:
            return order
        else:
            return order - self._n_features_y

    @property
    def undersampling(self):
        """Under-sampling threshold."""
        return np.floor(np.log2(self.n_samples))



"""
Functions to compute entropies.
"""

from functools import partial

import jax
import jax.numpy as jnp
import numpy as np
from jax.scipy.special import digamma as psi
from jax.scipy.special import gamma, ndtri
from jax.scipy.stats import gaussian_kde

from hoi.utils.logging import logger
from hoi.utils.stats import normalize

###############################################################################
###############################################################################
#                                 SWITCHER
###############################################################################
###############################################################################


def get_entropy(method="gc", **kwargs):
    """Get entropy function.

    Parameters
    ----------
    method : {'gc', 'gauss', 'binning', 'histogram', 'knn', 'kernel'}
        Name of the method to compute entropy. Use either :

            * 'gc': gaussian copula entropy [default]. See
                :func:`hoi.core.entropy_gc`
            * 'gauss': gaussian entropy. See :func:`hoi.core.entropy_gauss`
            * 'binning': estimator to use for discrete variables. See
                :func:`hoi.core.entropy_bin`
            * 'histogram' : estimator based on binning the data, to estimate
                the probability distribution of the variables and then
                compute the differential entropy. For more details see
                :func:`hoi.core.entropy_hist`
            * 'knn': k-nearest neighbor estimator. See
                :func:`hoi.core.entropy_knn`
            * 'kernel': kernel-based estimator of entropy
                see :func:`hoi.core.entropy_kernel`
            * A custom entropy estimator can be provided. It should be a
                callable function written with Jax taking a single 2D input
                of shape (n_features, n_samples) and returning a float.

    kwargs : dict | {}
        Additional arguments sent to the entropy function.

    Returns
    -------
    fcn : callable
        Function to compute entropy on a variable of shape
        (n_features, n_samples)
    """
    if method == "gc":
        return partial(entropy_gc, **kwargs)
    elif method == "gauss":
        return entropy_gauss
    elif method == "binning":
        return partial(entropy_bin, **kwargs)
    elif method == "histogram":
        return partial(entropy_hist, **kwargs)
    elif method == "knn":
        return partial(entropy_knn, **kwargs)
    elif method == "kernel":
        return partial(entropy_kernel, **kwargs)
    elif callable(method):
        # test the function
        try:
            assert method(np.random.rand(2, 100)).shape == ()
        except Exception:
            import traceback

            logger.error(traceback.format_exc())

            raise AssertionError(
                "A custom estimator should be a callable function written in "
                "Jax and taking as an input a variable x of shape (n_features,"
                " n_samples) and returning the entropy of the variable as a "
                "float."
            )

        # jit the function
        return jax.jit(method)
    else:
        raise ValueError(f"Method {method} doesn't exist.")


###############################################################################
###############################################################################
#                             PREPROCESSING
###############################################################################
###############################################################################


def prepare_for_it(data, method, samples=None, **kwargs):
    """Prepare the data before computing entropy."""
    # data.shape = n_variables, n_features, n_samples

    # -------------------------------------------------------------------------
    # type checking
    if (method in ["binning"]) and (data.dtype != int):
        raise ValueError(
            "data dtype should be integer. Check that you discretized your"
            " data. If so, use `data.astype(int)`"
        )
    elif (method in ["kernel", "gc", "knn", "histogram"]) and (
        data.dtype != float
    ):
        raise ValueError(f"data dtype should be float, not {data.dtype}")

    # -------------------------------------------------------------------------
    # trial selection
    if isinstance(samples, (np.ndarray, jnp.ndarray, list, tuple)):
        logger.info("    Sample selection")
        data = data[..., samples]
    data = jnp.asarray(data)

    # -------------------------------------------------------------------------
    # method specific preprocessing
    if method == "gc":
        logger.info("    Copnorm and demean the data")
        data = preproc_gc_3d(data)
        kwargs["copnorm"] = False
    # elif method == "kernel":
    #     logger.info("    Unit circle normalization")
    #     data = preproc_kernel_3d(data)

    return data, kwargs


###############################################################################
###############################################################################
#                              GAUSSIAN COPULA
###############################################################################
###############################################################################


@partial(jax.jit, static_argnums=(1, 2))
def entropy_gc(
    x: jnp.array, biascorrect: bool = True, copnorm: bool = True
) -> jnp.array:
    """Gaussian Copula entropy.

    Parameters
    ----------
    x : array_like
        Array of data of shape (n_features, n_samples)
    biascorrect : bool | True
        Specifies whether bias correction should be applied to the estimated MI
    copnorm : bool | True
        Apply gaussian copula normalization

    Returns
    -------
    hx : float
        Entropy of x (in bits)
    """
    nfeat, nsamp = x.shape

    # copula normalization
    if copnorm:
        x = preproc_gc_2d(x)

    # covariance
    c = jnp.dot(x, x.T) / float(nsamp - 1)
    chc = jnp.linalg.cholesky(c)

    # entropy in nats
    hx = jnp.sum(jnp.log(jnp.diagonal(chc))) + 0.5 * nfeat * (
        jnp.log(2 * jnp.pi) + 1.0
    )

    ln2 = jnp.log(2)
    if biascorrect:
        psiterms = (
            psi((nsamp - jnp.arange(1, nfeat + 1).astype(float)) / 2.0) / 2.0
        )
        dterm = (ln2 - jnp.log(nsamp - 1.0)) / 2.0
        hx = hx - nfeat * dterm - psiterms.sum()

    # convert to bits
    return hx / ln2


@jax.jit
def _preproc_gc(x: jnp.array) -> jnp.array:
    """Preprocessing for the Gaussian copula entropy and mutual-information.

    This function performs the two following steps on a vector :

        * Apply the copula normalization
        * Demean the copnormed data

    Parameters
    ----------
    x : array_like
        Array of data of shape (n_samples,)

    Returns
    -------
    cx : array_like
        Standard normal samples with the same empirical CDF value as the input.
    """
    # compute the empirical CDF
    cdf = jnp.argsort(jnp.argsort(x)).astype(float)
    cdf += 1.0
    cdf /= float(cdf.shape[-1] + 1)

    # infer normal distribution giving this CDF
    gauss = ndtri(cdf)

    # demean the gaussian
    gauss -= gauss.mean()

    return gauss


# preprocessing for a 2D variable (n_features, n_samples)
preproc_gc_2d = jax.jit(jax.vmap(_preproc_gc, in_axes=0))

# preprocessing for a 3D variable (n_variables, n_features, n_samples)
preproc_gc_3d = jax.jit(jax.vmap(preproc_gc_2d, in_axes=0))


###############################################################################
###############################################################################
#                                  GAUSSIAN
###############################################################################
###############################################################################


def entropy_gauss(x: jnp.array) -> jnp.array:
    """Gaussian entropy.

    Parameters
    ----------
    x : array_like
        Array of data of shape (n_features, n_samples)

    Returns
    -------
    hx : float
        Entropy of x (in bits)
    """
    return entropy_gc(x, biascorrect=False, copnorm=False)


###############################################################################
###############################################################################
#                               BINNING
###############################################################################
###############################################################################


@partial(jax.jit, static_argnums=(1,))
def entropy_bin(x: jnp.array, base: int = 2) -> jnp.array:
    """Entropy using binning.

    Parameters
    ----------
    x : array_like
        Input data of shape (n_features, n_samples). The data should already
        be discretize
    base : int | 2
        The logarithmic base to use. Default is base 2.

    Returns
    -------
    hx : float
        Entropy of x (in bits)
    """

    n_features, n_samples = x.shape

    # here, we count the number of possible multiplets. The worst is that each
    # trial is unique. So we can prepare the output to be at most (n_samples,)
    # and if trials are repeated, just set to zero it's going to be compensated
    # by the entr() function
    counts = jnp.unique(
        x, return_counts=True, size=n_samples, axis=1, fill_value=0
    )[1]
    probs = counts / n_samples

    return jax.scipy.special.entr(probs).sum() / jnp.log(base)


###############################################################################
###############################################################################
#                               HISTOGRAM
###############################################################################
###############################################################################


@partial(jax.jit, static_argnums=(1, 2))
def entropy_hist(x: jnp.array, base: float = 2, n_bins: int = 8) -> jnp.array:
    """Entropy using binning.

    Parameters
    ----------
    x : array_like
        Input data of shape (n_features, n_samples). The data should already
        be discretize
    base : float | 2
        The logarithmic base to use. Default is base 2.
    n_bins : int | 8
        The number of bin to be considered in the binarization process

    Returns
    -------
    hx : float
        Entropy of x (in bits)
    """

    # bin size computation
    bins_arr = (x.max(axis=1) - x.min(axis=1)) / n_bins
    bin_s = jnp.prod(bins_arr)

    # binning of the data
    x_min, x_max = x.min(axis=1, keepdims=True), x.max(axis=1, keepdims=True)
    dx = (x_max - x_min) / n_bins
    x_binned = ((x - x_min) / dx).astype(int)
    x_binned = jnp.minimum(x_binned, n_bins - 1).astype(int)

    n_features, n_samples = x_binned.shape

    # here, we count the number of possible multiplets. The worst is that each
    # trial is unique. So we can prepare the output to be at most (n_samples,)
    # and if trials are repeated, just set to zero it's going to be compensated
    # by the entr() function

    counts = jnp.unique(
        x_binned, return_counts=True, size=n_samples, axis=1, fill_value=0
    )[1]

    probs = counts / n_samples

    return bin_s * jax.scipy.special.entr(probs / bin_s).sum() / jnp.log(base)


###############################################################################
###############################################################################
#                                    KNN
###############################################################################
###############################################################################


@partial(jax.jit, static_argnums=(1,))
def entropy_knn(x, k: int = 3) -> jnp.array:
    """Entropy using the k-nearest neighbor.

    Original code: https://github.com/blakeaw/Python-knn-entropy/
    and references. See also Kraskov et al., Estimating mutual information,
    Phy rev, 2004

    Parameters
    ----------
    x : array_like
        Input data of shape (n_features, n_samples)
    knn : int | 1
        K-th closest point. Default is 1 (closest point)

    Returns
    -------
    hx : float
        Entropy of x (in bits)
    """
    # x = jnp.atleast_2d(x)
    d, n = float(x.shape[0]), float(x.shape[1])
    # compute euclidian distance
    x = x.T[None]
    diff = x.transpose(1, 0, 2) - x
    eucl_xi = jnp.sqrt(jnp.sum(diff**2, axis=-1))
    # dist to kth neighbor
    dist_k = jnp.sort(eucl_xi, axis=-1)[:, k]

    # volume of unit ball in d^n
    c_d = (jnp.pi ** (d * 0.5)) / gamma(1.0 + d * 0.5) / (2**d)
    log_c_d = jnp.log(c_d)

    # sum log of distances
    sum_log_dist = jnp.sum(jnp.log(2 * dist_k))

    h = -psi(k) + psi(n) + log_c_d + (d / n) * sum_log_dist

    return jnp.maximum(0, h) / jnp.log(2)  # added compared to original code


###############################################################################
###############################################################################
#                                  KERNEL
###############################################################################
###############################################################################


@partial(jax.jit, static_argnums=(1,))
def entropy_kernel(x: jnp.array, bw_method: str = None) -> jnp.array:
    """Entropy using gaussian kernel density.

    Parameters
    ----------
    x : array_like
        Input data of shape (n_features, n_samples)
    bw_method : str | None
        Estimator bandwidth. See jax.scipy.stats.gaussian_kde.

    Returns
    -------
    hx : float
        Entropy of x (in bits)
    """
    model = gaussian_kde(x, bw_method=bw_method)
    return -jnp.mean(jnp.log2(model(x)))
    # p = model.pdf(x)
    # return jax.scipy.special.entr(p).sum() / np.log(base)


# kernel preprocessing for a 2d variable (n_features, n_samples)
preproc_kernel_2d = jax.jit(
    jax.vmap(partial(normalize, to_min=-1.0, to_max=1.0), in_axes=0)
)

# kernel preprocessing for a 2d variable (n_variables, n_features, n_samples)
preproc_kernel_3d = jax.jit(jax.vmap(preproc_kernel_2d, in_axes=0))



import numpy as np

###############################################################################
###############################################################################
#                                 SWITCHER SIMULATIONS
###############################################################################
###############################################################################


def simulate_hoi_gauss(
    n_samples=1000,
    target=False,
    triplet_character="synergy",
):
    """Simulates High Order Interactions (HOIs) with or without target.

    This function can be used to simulate a triplet only using gaussian
    variables.

    Parameters
    ----------
    n_samples : int | 1000
        Number of samples to simulate.
    target : bool | False
        Indicates whether to include a target variable.
    triplet_character : {"redundancy", "synergy}
        Interaction character of the triplet of variables to generate. Choose
        either :

            * "redundancy" : a triplet of variables with positive O-information
              (about 0.3) will be generated
            * "synergy": a triplet of variables with negative O-information
              (about -0.8)

        In the case, target = True, triplet character refers to the information
        conveyed by the triplet with respect to the target variable.

    Returns
    -------
    data: np.ndarray
        Numpy array of simulated data of shape (n_samples, 3)
    target: np.ndarray
        Target variable of shape (n_samples,). This parameter is only returned
        when the input parameter target is set to True
    """
    assert triplet_character in ["redundancy", "synergy"]

    if not target:
        # without target
        return _sim_hoi(n_samples, triplet_character)

    elif target:
        # with target
        return _sim_hoi_target(n_samples, triplet_character)


###############################################################################
###############################################################################
#                                   STATIC HOIs
###############################################################################
###############################################################################


def _sim_hoi_target(n_samples, triplet_character):
    """Simulates High Order Interactions (HOIs) with target variable."""

    # Mean vector for multivariate Gaussian distribution with 4 variables
    mean_mvgauss = np.zeros(4)

    # Get the covariance matrix based on the triplet character
    cov = __cov_order_4(triplet_character)

    # Initialize an array to hold the simulated data
    simulated_data = np.zeros((n_samples, 4))

    # Generate the simulated data using a multivariate normal distribution
    simulated_data = np.random.multivariate_normal(
        mean_mvgauss, cov, size=n_samples, check_valid="warn", tol=1e-8
    )

    # Return the first three variables as the simulated data and the fourth as
    # # the target
    return simulated_data[:, :3], simulated_data[:, 3]


def _sim_hoi(n_samples, triplet_character):
    """Simulates High Order Interactions (HOIs) without target information."""

    # Mean vector for multivariate Gaussian distribution with 3 variables
    mean_mvgauss = np.zeros(3)

    # Get the covariance matrix based on the triplet character
    cov = __cov_order_3(triplet_character)

    # Initialize an array to hold the simulated data
    simulated_data = np.zeros((n_samples, 3))

    # Generate the simulated data using a multivariate normal distribution
    simulated_data = np.random.multivariate_normal(
        mean_mvgauss, cov, size=n_samples, check_valid="warn", tol=1e-8
    )

    # Return the simulated data
    return simulated_data


###############################################################################
###############################################################################
#                        COVARIANCE WITH HOI, ORDER 3 & 4
###############################################################################
###############################################################################


def __cov_order_3(character):
    """Compute the covariance matrix for three variables based
    on the given interaction character.

    Parameters
    ----------
    character : str
        'null', redundancy', or 'synergy' indicating the interaction
        character among the variables to simulate.

    Returns
    -------
    cov : numpy.ndarray
        Covariance matrix for the three variables with the specified
        interaction pattern.
    """

    # Define the standard deviations for the three variables
    lambx = np.sqrt(0.99)
    lamby = np.sqrt(0.7)
    lambz = np.sqrt(0.3)

    # Full factor matrix m
    m = np.array([lambx, lamby, lambz])[np.newaxis]

    if character == "redundancy":
        # We fix theta_yz in such a way that O(R1,R2,R3)>0
        theta_yz = 0.22

        # Noise covariances theta
        theta = np.diagflat(1 - m**2)
        theta += np.diagflat([0, theta_yz], 1) + np.diagflat([0, theta_yz], -1)

        # The covariance matrix for the three variables
        cov = m * m.T + theta

    elif character == "synergy":
        # We fix theta_yz in such a way that O(R1,R2,R3)<0
        theta_yz = -0.39

        # Noise covariances theta
        theta = np.diagflat(1 - m**2)
        theta += np.diagflat([0, theta_yz], 1) + np.diagflat([0, theta_yz], -1)

        # The covariance matrix for the three variables
        cov = m * m.T + theta

    return cov


def __cov_order_4(character):
    """Calculate the covariance matrix for a given interaction
    character.

    Parameters
    ----------
    character : str
        The character specifying the kind of interactions character
        of the three variables and the target variable. It can
        be either 'redundancy' or 'synergy'.

    Returns
    -------
    cov_ : numpy.ndarray
    The covariance matrix for the specified interaction character.
    """

    # Define the standard deviations for the four variables
    lambx = np.sqrt(0.99)
    lamby = np.sqrt(0.7)
    lambz = np.sqrt(0.3)
    lambs = np.sqrt(0.2)

    # Imposing the relationships with the target
    m = np.array([lambx, lamby, lambz, lambs])[np.newaxis]

    # Initialize the noise covariance matrix theta
    theta = np.diagflat(1 - m**2)

    if character == "redundancy":
        # We fix theta_zs in such a way that the variables are redundant
        theta_zs = 0.4

        # Update the noise covariance matrix theta
        theta += np.diagflat([0, 0, theta_zs], 1) + np.diagflat(
            [0, 0, theta_zs], -1
        )

        # Calculate the full covariance matrix
        cov_ = m * m.T + theta

    elif character == "synergy":
        # We fix theta_zs in such a way that the variables show synergy
        theta_zs = -0.75

        # Update the noise covariance matrix theta
        theta += np.diagflat([0, 0, theta_zs], 1) + np.diagflat(
            [0, 0, theta_zs], -1
        )

        # Calculate the full covariance matrix
        cov_ = m * m.T + theta

    return cov_



from functools import partial

import numpy as np

import jax
import jax.numpy as jnp

from hoi.metrics.base_hoi import HOIEstimator
from hoi.core.entropies import get_entropy, prepare_for_it
from hoi.utils.progressbar import get_pbar


@partial(jax.jit, static_argnums=(2, 3))
def _sinfo_no_ent(inputs, index, entropy_3d=None, entropy_4d=None):
    data, acc = inputs
    msize = len(index)

    # tuple selection
    x_c = data[:, index, :]

    # compute h(x^{n})
    h_xn = entropy_3d(x_c)

    # compute \sum_{j=1}^{n} h(x_{j}
    h_xj_sum = entropy_4d(x_c[:, :, jnp.newaxis, :]).sum(0)

    # compute \sum_{j=1}^{n} h(x_{-j}
    h_xmj_sum = entropy_4d(x_c[:, acc, :]).sum(0)

    # compute sinfo
    sinfo = -msize * h_xn + h_xj_sum + h_xmj_sum

    return inputs, sinfo


class Sinfo(HOIEstimator):
    r"""S-information.

    The S-information (also called exogenous information) is defined
    as the sum between the total correlation (TC) minus the dual total
    correlation (DTC):

    .. math::

        \Omega(X^{n})  &=  TC(X^{n}) + DTC(X^{n}) \\
                       &=  nH(X^{n}) + \sum_{j=1}^{n} [H(X_{j}) + H(
                        X_{-j}^{n})]

    Parameters
    ----------
    x : array_like
        Standard NumPy arrays of shape (n_samples, n_features) or
        (n_samples, n_features, n_variables)
    y : array_like
        The feature of shape (n_samples,) for estimating task-related S-info
    multiplets : list | None
        List of multiplets to compute. Should be a list of multiplets, for
        example [(0, 1, 2), (2, 7, 8, 9)]. By default, all multiplets are
        going to be computed.

    References
    ----------
    James et al., 2011 :cite:`james2011anatomy`
    """

    __name__ = "S-Information"
    _encoding = False
    _positive = "null"
    _negative = "null"
    _symmetric = True

    def __init__(self, x, y=None, multiplets=None, verbose=None):
        HOIEstimator.__init__(
            self, x=x, y=y, multiplets=multiplets, verbose=verbose
        )

    def fit(
        self, minsize=2, maxsize=None, method="gc", samples=None, **kwargs
    ):
        """Compute the S-information.

        Parameters
        ----------
        minsize, maxsize : int | 2, None
            Minimum and maximum size of the multiplets
        method : {'gc', 'binning', 'knn', 'kernel', callable}
            Name of the method to compute entropy. Use either :

                * 'gc': gaussian copula entropy [default]. See
                  :func:`hoi.core.entropy_gc`
                * 'gauss': gaussian entropy. See :func:`hoi.core.entropy_gauss`
                * 'binning': binning-based estimator of entropy. Note that to
                  use this estimator, the data have be to discretized. See
                  :func:`hoi.core.entropy_bin`
                * 'knn': k-nearest neighbor estimator. See
                  :func:`hoi.core.entropy_knn`
                * 'kernel': kernel-based estimator of entropy
                  see :func:`hoi.core.entropy_kernel`
                * A custom entropy estimator can be provided. It should be a
                  callable function written with Jax taking a single 2D input
                  of shape (n_features, n_samples) and returning a float.

        samples : np.ndarray
            List of samples to use to compute HOI. If None, all samples are
            going to be used.
        kwargs : dict | {}
            Additional arguments are sent to each entropy function

        Returns
        -------
        hoi : array_like
            The NumPy array containing values of higher-order interactions of
            shape (n_multiplets, n_variables)
        """
        # ________________________________ I/O ________________________________
        # check min and max sizes
        minsize, maxsize = self._check_minmax(minsize, maxsize)

        # prepare the x for computing entropy
        x, kwargs = prepare_for_it(self._x, method, samples=samples, **kwargs)

        # get entropy function
        entropy = jax.vmap(get_entropy(method=method, **kwargs))
        sinfo_no_ent = partial(
            _sinfo_no_ent,
            entropy_3d=entropy,
            entropy_4d=jax.vmap(entropy, in_axes=1),
        )

        # get multiplet indices and order
        h_idx, order = self.get_combinations(minsize, maxsize=maxsize)

        # get progress bar
        pbar = get_pbar(
            iterable=range(order.min(), order.max() + 1), leave=False
        )

        # ______________________________ ENTROPY ____________________________
        offset = 0
        hoi = jnp.zeros((len(order), self.n_variables), dtype=jnp.float32)
        for msize in pbar:
            pbar.set_description(desc="Sinfo (%i)" % msize, refresh=False)

            # get the number of features when considering y
            n_feat_xy = msize + self._n_features_y

            # combinations of features
            _h_idx = h_idx[order == msize, 0:n_feat_xy]

            # indices for X_{-j} and skip first column
            acc = jnp.mgrid[0:n_feat_xy, 0:n_feat_xy].sum(0) % n_feat_xy
            acc = acc[:, 1:]

            # compute sinfo
            _, _hoi = jax.lax.scan(sinfo_no_ent, (x, acc), _h_idx)

            # fill variables
            n_combs = _h_idx.shape[0]
            hoi = hoi.at[offset : offset + n_combs, :].set(_hoi)

            # updates
            offset += n_combs

        return np.asarray(hoi)



"""
"""

import logging
import sys
import re

BLACK, RED, GREEN, YELLOW, BLUE, MAGENTA, CYAN, WHITE = range(8)
RESET_SEQ = "\033[0m"
COLOR_SEQ = "\033[1;%dm"
BOLD_SEQ = "\033[1m"
COLORS = {
    "DEBUG": GREEN,
    "PROFILER": MAGENTA,
    "INFO": WHITE,
    "WARNING": YELLOW,
    "ERROR": RED,
    "CRITICAL": RED,
}
FORMAT = {
    "compact": "$BOLD%(levelname)s | %(message)s",
    "spacy": "$BOLD%(levelname)-19s$RESET | %(message)s",
    "frites": "$BOLD%(name)s-%(levelname)-19s$RESET | %(message)s",
    "print": "%(message)s",
}
_LOGGING_TYPES = dict(
    DEBUG=logging.DEBUG,
    INFO=logging.INFO,
    WARNING=logging.WARNING,
    ERROR=logging.ERROR,
    CRITICAL=logging.CRITICAL,
)


def formatter_message(message, use_color=True):
    """Format the message."""
    return message.replace("$RESET", RESET_SEQ).replace("$BOLD", BOLD_SEQ)


class _Formatter(logging.Formatter):
    """Formatter."""

    def __init__(self, format_type="compact"):
        logging.Formatter.__init__(self, FORMAT[format_type])
        self._format_type = format_type

    def format(self, record):
        name = record.levelname
        msg = record.getMessage()
        # If * in msg, set it in RED :
        if "*" in msg:
            regexp = "\*.*?\*"
            re_search = re.search(regexp, msg).group()
            to_color = (
                COLOR_SEQ % (30 + RED)
                + re_search
                + COLOR_SEQ % (30 + WHITE)
                + RESET_SEQ
            )
            msg_color = re.sub(regexp, to_color, msg)
            msg_color += RESET_SEQ
            record.msg = msg_color
        # Set level color :
        levelname_color = COLOR_SEQ % (30 + COLORS[name]) + name + RESET_SEQ
        record.levelname = levelname_color
        if record.levelno == 20:
            logging.Formatter.__init__(self, FORMAT["print"])
        else:
            logging.Formatter.__init__(self, FORMAT[self._format_type])
        return formatter_message(logging.Formatter.format(self, record))


class _StreamHandler(logging.StreamHandler):
    """Stream handler allowing matching and recording."""

    def __init__(self):
        logging.StreamHandler.__init__(self, sys.stderr)
        self.setFormatter(_lf)
        self._str_pattern = None
        self.emit = self._frites_emit

    def _frites_emit(self, record, *args):
        msg = record.getMessage()
        test = self._match_pattern(record, msg)
        if test:
            record.msg = test
            return logging.StreamHandler.emit(self, record)
        else:
            return

    def _match_pattern(self, record, message):
        if isinstance(self._str_pattern, str):
            if re.search(self._str_pattern, message):
                sub = "*{}*".format(self._str_pattern)
                return re.sub(self._str_pattern, sub, message)
            else:
                return ""
        else:
            return message


logger = logging.getLogger("hoi")
# logger.propagate = True
_lf = _Formatter()
_lh = _StreamHandler()  # needs _lf to exist
logger.addHandler(_lh)
PROFILER_LEVEL_NUM = 1
logging.addLevelName(PROFILER_LEVEL_NUM, "PROFILER")


def profiler_fcn(self, message, *args, **kws):  # noqa
    # Yes, logger takes its '*args' as 'args'.
    if self.isEnabledFor(PROFILER_LEVEL_NUM):
        self._log(PROFILER_LEVEL_NUM, message, args, **kws)


logging.Logger.profiler = profiler_fcn
logger.setLevel(logging.INFO)


def set_log_level(verbose=None, return_old_level=False):
    """Set the logging level.

    Parameters
    ----------
    verbose : bool, str, int, or None
        The verbosity of messages to print. If a str, it can be either DEBUG,
        INFO, WARNING, ERROR, or CRITICAL. Note that these are for
        convenience and are equivalent to passing in logging.DEBUG, etc.
        For bool, True is the same as 'INFO', False is the same as 'WARNING'.
        If None, the environment variable MNE_LOGGING_LEVEL is read, and if
        it doesn't exist, defaults to INFO.
    return_old_level : bool
        If True, return the old verbosity level.

    Returns
    -------
    old_level : int
        The old level. Only returned if ``return_old_level`` is True.
    """
    old_verbose = logger.level

    if verbose is None:
        verbose = "INFO"
    elif isinstance(verbose, bool):
        if verbose is True:
            verbose = "INFO"
        else:
            verbose = "WARNING"
    if isinstance(verbose, str):
        verbose = verbose.upper()
        verbose = _LOGGING_TYPES[verbose]

    if verbose != old_verbose:
        logger.setLevel(verbose)
    return old_verbose if return_old_level else None



from tqdm.auto import tqdm
from hoi.utils.logging import logger


def get_pbar(**kwargs):
    """Get progress bar"""
    kwargs["disable"] = logger.getEffectiveLevel() > 20
    kwargs["mininterval"] = 0.016
    kwargs["miniters"] = 1
    kwargs["smoothing"] = 0.05
    kwargs["bar_format"] = (
        "{percentage:3.0f}%|{bar}| {desc} {n_fmt}/{total_fmt} [{elapsed}"
        "<{remaining}, {rate_fmt:>11}{postfix}]"
    )
    return tqdm(**kwargs)



import warnings

import numpy as np

import jax
import jax.numpy as jnp

from hoi.metrics.base_hoi import HOIEstimator
from jax_tqdm import scan_tqdm


###############################################################################
###############################################################################
#                            MUTUAL INFORMATION
###############################################################################
###############################################################################


# @jax.jit
# def find_indices(inputs, c):
#     combs, keep = inputs

#     keep = jnp.add(keep, (combs == c).any(1).astype(int))

#     return (combs, keep), None


@jax.jit
def compute_mi(inputs, iterators):
    combs, h, order = inputs
    _, comb = iterators

    # scanning over indices
    # is_inside = jnp.zeros((combs.shape[0],), dtype=int)
    # (_, is_inside), _ = jax.lax.scan(find_indices, (combs, is_inside), comb)

    # tensor implementation
    is_inside = (combs == comb[jnp.newaxis, ...]).sum((1, 2))

    # compute infotopo
    info = jnp.sum(h, where=(is_inside == order).reshape(-1, 1), axis=0)

    return inputs, info


class InfoTopo(HOIEstimator):
    """Topological Information.

    The multivariate mutual information :math:`I_{k}` quantify the
    variability/randomness and the statistical dependences between variables
    is defined by :

    .. math::

        I_{k}(X_{1}; ...; X_{k}) = \sum_{i=1}^{k} (-1)^{i - 1} \sum_{
         I\subset[k];card(I)=i} H_{i}(X_{I})

    .. warning::

        * :math:`I_{k}(X_{1}; ...; X_{k}) > 0 \Rightarrow Redundancy`
        * :math:`I_{k}(X_{1}; ...; X_{k}) < 0 \Rightarrow Synergy`

    Parameters
    ----------
    x : array_like
        Standard NumPy arrays of shape (n_samples, n_features) or
        (n_samples, n_features, n_variables)

    References
    ----------
    Baudot et al., 2019 :cite:`baudot2019infotopo`; Tapia et al., 2018
    :cite:`tapia2018neurotransmitter`
    """

    __name__ = "Topological Information"
    _encoding = False
    _positive = "redundancy"
    _negative = "synergy"
    _symmetric = True

    def __init__(self, x, y=None, verbose=None):
        # for infotopo, the multiplets are set to None because this metric
        # first require to compute entropies and then associate them to form
        # the MI. Same for the target y.
        if y is not None:
            warnings.warn("For InfoTopo, y input is going to be ignored.")
        HOIEstimator.__init__(
            self, x=x, y=None, multiplets=None, verbose=verbose
        )

    def fit(
        self, minsize=1, maxsize=None, method="gc", samples=None, **kwargs
    ):
        """Compute Topological Information.

        Parameters
        ----------
        minsize, maxsize : int | 2, None
            Minimum and maximum size of the multiplets
        method : {'gc', 'binning', 'knn', 'kernel', callable}
            Name of the method to compute entropy. Use either :

                * 'gc': gaussian copula entropy [default]. See
                  :func:`hoi.core.entropy_gc`
                * 'gauss': gaussian entropy. See :func:`hoi.core.entropy_gauss`
                * 'binning': binning-based estimator of entropy. Note that to
                  use this estimator, the data have be to discretized. See
                  :func:`hoi.core.entropy_bin`
                * 'knn': k-nearest neighbor estimator. See
                  :func:`hoi.core.entropy_knn`
                * 'kernel': kernel-based estimator of entropy
                  see :func:`hoi.core.entropy_kernel`
                * A custom entropy estimator can be provided. It should be a
                  callable function written with Jax taking a single 2D input
                  of shape (n_features, n_samples) and returning a float.

        samples : np.ndarray
            List of samples to use to compute HOI. If None, all samples are
            going to be used.
        kwargs : dict | {}
            Additional arguments are sent to each entropy function

        Returns
        -------
        hoi : array_like
            The NumPy array containing values of higher-order interactions of
            shape (n_multiplets, n_variables)
        """
        # ____________________________ ENTROPIES ______________________________

        minsize, maxsize = self._check_minmax(minsize, maxsize)
        h_x, h_idx, order = self.compute_entropies(
            minsize=1,
            maxsize=maxsize,
            method=method,
            samples=samples,
            **kwargs,
        )

        # _______________________________ HOI _________________________________

        # compute order and multiply entropies
        h_x_sgn = jnp.multiply(((-1.0) ** (order.reshape(-1, 1) - 1)), h_x)

        # subselection of multiplets
        mults, _ = self.get_combinations(minsize, maxsize=maxsize)
        h_idx_2 = jnp.where(mults == -1, -2, mults)
        n_mult = mults.shape[0]

        # progress-bar definition
        pbar = scan_tqdm(n_mult, message="Mutual information")

        # compute mi
        _, hoi = jax.lax.scan(
            pbar(compute_mi),
            (h_idx[..., jnp.newaxis], h_x_sgn, order),
            (jnp.arange(n_mult), h_idx_2),
        )

        return np.asarray(hoi)



from .combinatory import combinations  # noqa
from .entropies import (  # noqa
    get_entropy,
    entropy_gc,
    entropy_gauss,
    entropy_bin,
    entropy_hist,
    entropy_knn,
    prepare_for_it,
    entropy_kernel,
)
from .mi import get_mi  # noqa



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# License: 3-clause BSD
import os
from setuptools import setup, find_packages

__version__ = "0.0.4"
NAME = "hoi"
AUTHOR = "BraiNets"
MAINTAINER = "Etienne Combrisson"
EMAIL = "e.combrisson@gmail.com"
KEYWORDS = "information-theory statistics higher-order-interactions"
DESCRIPTION = "Higher Order Interactions"
URL = "https://github.com/brainets/hoi"
DOWNLOAD_URL = (
    "https://github.com/brainets/hoi/archive/v" + __version__ + ".tar.gz"
)
# Data path :
PACKAGE_DATA = {}


def read(fname):
    """Read README and LICENSE."""
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


with open("requirements.txt") as f:
    requirements = f.read().splitlines()

test_deps = ["pytest", "pytest-cov", "codecov", "xarray"]
doc_deps = [
    "sphinx!=4.1.0",
    "sphinx-gallery",
    "sphinx_design",
    "sphinx-book-theme",
    "sphinxcontrib-bibtex",
    "sphinx_contributors",
    "numpydoc",
    "matplotlib",
    "xgi",
]
lint_deps = ["flake8", "pep8-naming", "black"]

setup(
    name=NAME,
    version=__version__,
    packages=find_packages(),
    package_dir={"hoi": "hoi"},
    package_data=PACKAGE_DATA,
    include_package_data=True,
    description=DESCRIPTION,
    long_description=read("README.rst"),
    platforms="any",
    setup_requires=["numpy"],
    install_requires=requirements,
    extras_require={
        "test": requirements + test_deps,
        "doc": requirements + doc_deps,
        "lint": requirements + lint_deps,
        "full": requirements + test_deps + doc_deps + lint_deps,
    },
    dependency_links=[],
    author=AUTHOR,
    maintainer=MAINTAINER,
    author_email=EMAIL,
    url=URL,
    download_url=DOWNLOAD_URL,
    license="BSD 3-Clause License",
    keywords=KEYWORDS,
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Education",
        "Intended Audience :: Developers",
        "Topic :: Scientific/Engineering :: Visualization",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
    ],
)



import numpy as np

from hoi.utils import normalize, landscape


def plot_landscape(
    hoi,
    model=None,
    orders=None,
    minsize=None,
    maxsize=None,
    kind="hist",
    undersampling=True,
    cbar=True,
    hist_kwargs={},
    plt_kwargs={},
):
    """Landscape representation of higher order interactions.

    Parameters
    ----------
    hoi : array_like
        Higher order interactions.
    model : hoi.metrics
        Higher order interaction model.
    orders: array_like, optional
        Order associated to each multiplet. The default is None.
    minsize, maxsize : int | 2, None
        Minimum and maximum size of the multiplets
    kind : {'hist', 'scatter'}
        Kind of plot. Use either:

            * 'hist' : 2D histogram of the higher order interactions
            * 'scatter' : scatter plot of the higher order interactions

    undersampling : bool | True
        If True, plot the undersampling threshold.
    cbar : bool | True
        Add colorbar.
    hist_kwargs : dict | {}
        Optional arguments for the histogram.
    plt_kwargs : dict | {}
        Optional arguments for the plot. If kind is 'hist', the arguments
        are passed to `plt.pcolormesh`. If kind is 'scatter', the arguments
        are passed to `plt.scatter`.

    Returns
    -------
    ax : axis
        Current axis.
    """
    import matplotlib.pyplot as plt
    from matplotlib.colors import LogNorm

    # get orders
    if model:
        orders = model.order

    # get computed orders
    if minsize is None:
        minsize = orders.min()
    if maxsize is None:
        maxsize = orders.max()

    # order selection
    keep = np.logical_and(orders >= minsize, orders <= maxsize)
    hoi = hoi[keep]
    orders = orders[keep]

    # switch between histogram and scatter plot
    if kind == "hist":
        hist_kwargs["output"] = "numpy"
        lscp, order, bins = landscape(hoi, orders, **hist_kwargs)

        plt.pcolormesh(order, bins, lscp, norm=LogNorm(), **plt_kwargs)
        if cbar:
            plt.colorbar(label=hist_kwargs.get("stat", "probability"))
    elif kind == "scatter":
        size = normalize(np.abs(hoi), to_min=2, to_max=200)
        minmax = abs(np.nanpercentile(hoi, [1, 99])).max()
        for o in range(minsize, maxsize + 1):
            keep = orders == o
            hoi_o = hoi[keep]
            x = np.random.normal(loc=o, scale=0.13, size=hoi_o.size)
            plt.scatter(
                x,
                hoi_o,
                c=hoi_o,
                vmin=-minmax,
                vmax=minmax,
                s=size[keep],
                **plt_kwargs,
            )

    plt.xlabel("Order")
    plt.xticks(np.arange(minsize, maxsize + 1))
    if model:
        plt.ylabel(f"{model.__name__} [Bits]")
    else:
        plt.ylabel("HOI [Bits]")
    plt.grid(True)

    if undersampling and model:
        plt.axvline(model.undersampling, color="k", linestyle="--")

    return plt.gca()



from functools import partial

import jax
import jax.numpy as jnp
import numpy as np
from jax.scipy.special import digamma as psi

from hoi.core.entropies import get_entropy, preproc_gc_2d
from hoi.utils.logging import logger

###############################################################################
###############################################################################
#                                 SWITCHER
###############################################################################
###############################################################################


def get_mi(method="gc", **kwargs):
    """Get Mutual-Information function.

    Parameters
    ----------
    method : {'gc', 'gauss', 'binning', 'knn', 'kernel'}
        Name of the method to compute mutual-information.
    kwargs : dict | {}
        Additional arguments sent to the mutual-information function.

    Returns
    -------
    fcn : callable
        Function to compute mutual information on variables of shapes
        (n_features, n_samples)
    """
    if method == "gc":
        return partial(mi_gc, **kwargs)
    elif method == "gauss":
        return mi_gauss
    elif method == "knn":
        return partial(mi_knn, **kwargs)
    elif callable(method):
        # test the function
        try:
            x = np.random.rand(2, 100)
            y = np.random.rand(4, 100)
            assert method(x, y).shape == ()
        except Exception:
            import traceback

            logger.error(traceback.format_exc())

            raise AssertionError(
                "A custom estimator should be a callable function written in "
                "Jax and taking two inputs x and y of shapes (n_features_x,"
                " n_samples) and (n_features_y, n_samples) and returning the "
                "mutual information between variables as a float."
            )

        # jit the function
        return jax.jit(method)
    else:
        # get the entropy function
        _entropy = get_entropy(method=method, **kwargs)

        # wrap the mi function with it
        return partial(compute_mi, entropy_fcn=_entropy)


###############################################################################
###############################################################################
#                             PREPROCESSING
###############################################################################
###############################################################################


# def prepare_for_mi(x, y, method, **kwargs):
#     """Prepare the data before computing mutual-information."""
#     x, kwargs = prepare_for_it(x, method, **kwargs.copy())
#     return x, y, kwargs


@partial(jax.jit, static_argnums=(2))
def compute_mi_comb(inputs, comb, mi=None):
    x, y = inputs
    x_c = x[:, comb, :]
    return inputs, mi(x_c, y)


@partial(jax.jit, static_argnums=(2))
def compute_mi_comb_phi(inputs, comb, mi=None):
    x, y = inputs
    x_c = jnp.atleast_2d(x[:, comb[0], :])
    y_c = y[:, comb[1], :]

    return inputs, mi(x_c, y_c)


###############################################################################
###############################################################################
#                         GENERAL MUTUAL INFORMATION
###############################################################################
###############################################################################


@partial(jax.jit, static_argnums=(2,))
def compute_mi(x, y, entropy_fcn=None):
    """Compute the mutual-information by providing an entropy function.

    Parameters
    ----------
    x, y : array_like
        Arrays to consider for computing the Mutual Information. The two input
        variables x and y should have a shape of (n_features_x, n_samples) and
        (n_features_y, n_samples)
    entropy_fcn : function | None
        Function to use for computing the entropy.

    Returns
    -------
    mi : float
        Floating value describing the mutual-information between x and y
    """
    # compute mi
    mi = (
        entropy_fcn(x)
        + entropy_fcn(y)
        - entropy_fcn(jnp.concatenate((x, y), axis=0))
    )
    return mi


###############################################################################
###############################################################################
#                     GAUSSIAN COPULA MUTUAL INFORMATION
###############################################################################
###############################################################################


@partial(jax.jit, static_argnums=(2, 3))
def mi_gc(
    x: jnp.array,
    y: jnp.array,
    biascorrect: bool = False,
    copnorm: bool = False,
):
    """Mutual information (MI) between two Gaussian variables in bits.

    Parameters
    ----------
    x, y : array_like
        Arrays to consider for computing the Mutual Information. The two input
        variables x and y should have a shape of (n_features_x, n_samples) and
        (n_features_y, n_samples)
    biascorrect : bool | False
        Specifies whether bias correction should be applied to the estimated MI
    copnorm : bool | True
        Apply gaussian copula normalization


    Returns
    -------
    i : float
        Information shared by x and y (in bits)
    """
    n_features_x, n_samples = x.shape
    n_features_y = y.shape[0]
    n_features_xy = n_features_x + n_features_y

    # joint variable
    xy = jnp.vstack((x, y))
    if copnorm:
        xy = preproc_gc_2d(xy)
    cxy = jnp.dot(xy, xy.T) / float(n_samples - 1)
    # submatrices of joint covariance
    cx = cxy[:n_features_x, :n_features_x]
    cy = cxy[n_features_x:, n_features_x:]

    chcxy = jnp.linalg.cholesky(cxy)
    chcx = jnp.linalg.cholesky(cx)
    chcy = jnp.linalg.cholesky(cy)

    # entropies in nats
    # normalizations cancel for mutual information
    hx = jnp.sum(jnp.log(jnp.diagonal(chcx)))
    hy = jnp.sum(jnp.log(jnp.diagonal(chcy)))
    hxy = jnp.sum(jnp.log(jnp.diagonal(chcxy)))

    ln2 = jnp.log(2)
    if biascorrect:
        psiterms = (
            psi(
                (n_samples - jnp.arange(1, n_features_xy + 1)).astype(float)
                / 2.0
            )
            / 2.0
        )
        dterm = (ln2 - jnp.log(n_samples - 1.0)) / 2.0
        hx = hx - n_features_x * dterm - psiterms[:n_features_x].sum()
        hy = hy - n_features_y * dterm - psiterms[:n_features_y].sum()
        hxy = hxy - n_features_xy * dterm - psiterms[:n_features_xy].sum()

    # MI in bits
    i = (hx + hy - hxy) / ln2
    return i


###############################################################################
###############################################################################
#                       GAUSSIAN MUTUAL INFORMATION
###############################################################################
###############################################################################


def mi_gauss(x: jnp.array, y: jnp.array):
    """Mutual information (MI) between two Gaussian variables in bits.

    Parameters
    ----------
    x, y : array_like
        Arrays to consider for computing the Mutual Information. The two input
        variables x and y should have a shape of (n_features_x, n_samples) and
        (n_features_y, n_samples)

    Returns
    -------
    i : float
        Information shared by x and y (in bits)
    """
    return mi_gc(x, y, biascorrect=False, copnorm=False)


###############################################################################
###############################################################################
#                         KNN MUTUAL INFORMATION
###############################################################################
###############################################################################


@jax.jit
def _cdist(x, y) -> jnp.ndarray:
    """Pairwise squared distances between all samples of x and y."""
    diff = x.T[:, None, :] - y.T[None]
    _dist = jnp.sum(diff**2, axis=-1)
    return _dist


@partial(jax.jit, static_argnums=(2,))
def mi_knn(x, y, k: int = 3) -> jnp.array:
    """Mutual information using the KSG estimator.

    First algorithm proposed in Kraskov et al., Estimating mutual information,
    Phy rev, 2004.

    Parameters
    ----------
    x, y : array_like
        Arrays to consider for computing the Mutual Information. The two input
        variables x and y should have a shape of (n_features_x, n_samples) and
        (n_features_y, n_samples)
    k : int
        Number of nearest neighbors to consider for the KSG estimator.

    Returns
    -------
    mi : float
        Floating value describing the mutual-information between x and y.
    """
    # n_samples
    n = float(x.shape[1])
    # for each xi and yi, get the distance to neighbors
    eucl_xi = jnp.sqrt(_cdist(x, x))
    eucl_yi = jnp.sqrt(_cdist(y, y))

    # distance in space (XxY) is the maximum distance.
    max_dist_xy = jnp.maximum(eucl_xi, eucl_yi)
    # indices to the closest points in the (XxY) space.
    closest_points = jnp.argsort(max_dist_xy, axis=-1)
    # the kth neighbour is at index k (ignoring the point itself)
    # distance to the k-th neighbor for each point
    k_neighbours = closest_points[:, k]
    dist_k = max_dist_xy[jnp.arange(len(k_neighbours)), k_neighbours]
    # don't include the `i`th point itself in nx and ny
    nx = (eucl_xi < dist_k[:, None]).sum(axis=1) - 1
    ny = (eucl_yi < dist_k[:, None]).sum(axis=1) - 1

    psi_mean = jnp.sum((psi(nx + 1) + psi(ny + 1)) / n)

    mi = psi(k) - psi_mean + psi(n)
    return mi / jnp.log(2)



from hoi.metrics.oinfo import Oinfo
from hoi.metrics.base_hoi import HOIEstimator


class GradientOinfo(HOIEstimator):
    r"""First order Gradient O-information.

    The Gradient O-information is defined as the difference between the
    O-information with the target variable minus the O-information without the
    target variable :

    .. math::

        \partial_{i}\Omega(X^{n}) &= \Omega(X^{n}) - \Omega(X^{n}_{-i}) \\
                                  &= (2 - n)I(X_{i}; X^{n}_{-i}) + \sum_{
                                   k=1, k\neq i}^{n} I(X_{k}; X^{n}_{-ik})

    .. warning::

        * :math:`\partial_{i}\Omega(X^{n}) > 0 \Rightarrow Redundancy`
        * :math:`\partial_{i}\Omega(X^{n}) < 0 \Rightarrow Synergy`

    Parameters
    ----------
    x : array_like
        Standard NumPy arrays of shape (n_samples, n_features) or
        (n_samples, n_features, n_variables)
    y : array_like
        The feature of shape (n_samples,) for estimating task-related O-info
    multiplets : list | None
        List of multiplets to compute. Should be a list of multiplets, for
        example [(0, 1, 2), (2, 7, 8, 9)]. By default, all multiplets are
        going to be computed.

    References
    ----------
    Scagliarini et al., 2023 :cite:`scagliarini2023gradients`
    """

    __name__ = "Gradient O-Information"
    _encoding = True
    _positive = "redundancy"
    _negative = "synergy"
    _symmetric = True

    def __init__(self, x, y, multiplets=None, base_model=Oinfo, verbose=None):
        kw_oinfo = dict(multiplets=multiplets, verbose=verbose)
        HOIEstimator.__init__(self, x=x, y=None, **kw_oinfo)
        self._model_tr = base_model(x, y=y, **kw_oinfo)
        self._model_tf = base_model(x, **kw_oinfo)
        self.__name__ = self.__name__ + "(%s)" % base_model.__name__

    def fit(
        self, minsize=2, maxsize=None, method="gc", samples=None, **kwargs
    ):
        """Compute the Gradient O-information.

        Parameters
        ----------
        minsize, maxsize : int | 2, None
            Minimum and maximum size of the multiplets
        method : {'gc', 'binning', 'knn', 'kernel', callable}
            Name of the method to compute entropy. Use either :

                * 'gc': gaussian copula entropy [default]. See
                  :func:`hoi.core.entropy_gc`
                * 'gauss': gaussian entropy. See :func:`hoi.core.entropy_gauss`
                * 'binning': binning-based estimator of entropy. Note that to
                  use this estimator, the data have be to discretized. See
                  :func:`hoi.core.entropy_bin`
                * 'knn': k-nearest neighbor estimator. See
                  :func:`hoi.core.entropy_knn`
                * 'kernel': kernel-based estimator of entropy
                  see :func:`hoi.core.entropy_kernel`
                * A custom entropy estimator can be provided. It should be a
                  callable function written with Jax taking a single 2D input
                  of shape (n_features, n_samples) and returning a float.

        samples : np.ndarray
            List of samples to use to compute HOI. If None, all samples are
            going to be used.
        kwargs : dict | {}
            Additional arguments are sent to each entropy function

        Returns
        -------
        hoi : array_like
            The NumPy array containing values of higher-order interactions of
            shape (n_multiplets, n_variables)
        """
        kw_oinfo = dict(
            minsize=minsize,
            maxsize=maxsize,
            method=method,
            samples=samples,
            **kwargs,
        )

        # ____________________________ TASK-FREE ______________________________
        hoi_tf = self._model_tf.fit(**kw_oinfo)

        self._multiplets = self._model_tf._multiplets

        # __________________________ TASK-RELATED _____________________________
        hoi_tr = self._model_tr.fit(**kw_oinfo)

        return hoi_tr - hoi_tf



"""Statistics and summary statistics on HOI.
"""

from functools import partial

import jax
import jax.numpy as jnp
import numpy as np


def landscape(
    x,
    mult_size,
    n_bins=100,
    centered=False,
    stat="probability",
    output="numpy",
):
    """Compute the landscape from HOI values.

    The landscape represents the how estimates of HOI are distributed per
    order.

    Parameters
    ----------
    x : array_like
        Array of data containing the HOI estimates of shape (n_hoi,)
    multi_size : array_like
        Size of the multiplet associated to each HOI estimates. It should be an
        array of shape (n_hoi,)
    n_bins : array_like | 100
        Number of bins to use to build the histogram at each order
    centered : bool | False
        Specify whether bin edges should be centered around zero
    stat : {'probability', 'frequency', 'count', 'density', 'percent'}
        Aggregate statistic to compute in each bin.

            * count: show the number of observations in each bin
            * frequency: show the number of observations divided by the bin
              width
            * probability or proportion: normalize such that bar heights sum to
              1
            * percent: normalize such that bar heights sum to 100
            * density: normalize such that the total area of the histogram
              equals 1
    output : {'numpy', 'pandas', 'xarray'} | None
        Output type. Use either :

            * 'numpy': in that case this function will returns the landscape,
              the unique orders (x-axis) and the central bins (y-axis)
            * 'pandas': 2-dimensional pandas dataframe
            * 'xarray': 2-dimensional dataarray

    Returns
    -------
    landcape : array_like
        Returns depend on the `output` parameter. Check to see what is returned
    """
    assert len(x) == len(mult_size)
    assert stat in [
        "probability",
        "proportion",
        "frequency",
        "count",
        "density",
        "percent",
    ]

    # get multiplet size
    mult_size = np.asarray(mult_size)
    minsize, maxsize = np.min(mult_size), np.max(mult_size)
    n_orders = maxsize - minsize + 1
    msize = np.arange(minsize, maxsize + 1)

    # define bin edges
    omin, omax = np.nanmin(x), np.nanmax(x)
    if centered:
        minmax = np.nanmax(np.abs([omin, omax]))
        o_range = (-minmax, minmax)
    else:
        o_range = (omin, omax)
    edges = np.histogram_bin_edges(x, bins=n_bins, range=o_range)
    edge_centers = (edges[0:-1] + edges[1::]) / 2.0

    # compute histogram
    lscp = np.zeros((n_bins, n_orders))
    for n_m, mult in enumerate(msize):
        idx = np.where(mult_size == mult)[0]
        _hist = np.histogram(x[idx], bins=edges, density=stat == "density")[0]

        if stat in ["probability", "proportion"]:
            _hist = _hist.astype(float) / np.sum(_hist)
        elif stat == "percent":
            _hist = 100.0 * _hist.astype(float) / np.sum(_hist)
        elif stat == "frequency":
            _hist = _hist.astype(float) / np.diff(edges)

        lscp[:, n_m] = _hist
    lscp[lscp == 0.0] = np.nan

    # output type
    if output == "numpy":
        return lscp, msize, edge_centers
    elif output == "pandas":
        import pandas as pd

        return pd.DataFrame(
            np.flipud(lscp), columns=msize, index=edge_centers[::-1]
        )
    elif output == "xarray":
        import xarray as xr

        attrs = dict(stat=stat, n_bins=n_bins)
        lscp = xr.DataArray(
            lscp,
            dims=("bins", "order"),
            coords=(edge_centers, msize),
            name=stat,
            attrs=attrs,
        )
        lscp.bins.attrs["unit"] = "bits"
        return lscp

    return lscp


def digitize_1d(x, n_bins):
    """One dimensional digitization."""
    assert x.ndim == 1
    x_min, x_max = x.min(), x.max()
    dx = (x_max - x_min) / n_bins
    x_binned = ((x - x_min) / dx).astype(int)
    x_binned = np.minimum(x_binned, n_bins - 1)
    return x_binned.astype(int)


def digitize_sklearn(x, **kwargs):
    """One dimensional digitization."""
    assert x.ndim == 1
    from sklearn.preprocessing import KBinsDiscretizer

    return (
        KBinsDiscretizer(**kwargs)
        .fit_transform(x.reshape(-1, 1))
        .astype(int)
        .squeeze()
    )


def digitize(x, n_bins, axis=0, use_sklearn=False, **kwargs):
    """Discretize a continuous variable.

    Parameters
    ----------
    x : array_like
        Array to discretize
    n_bins : int
        Number of bins
    axis : int | 0
        Axis along which to perform the discretization. By default,
        discretization is performed along the first axis (n_samples,)
    use_sklearn : bool | False
        If True, use sklearn.preprocessing.KBinsDiscretizer to discretize the
        data.
    kwargs : dict | {}
        Additional arguments are passed to
        sklearn.preprocessing.KBinsDiscretizer. For example, use
        `strategy='quantile'` for equal population binning.

    Returns
    -------
    x_binned : array_like
        Digitized array with the same shape as x
    """
    if not use_sklearn:
        return np.apply_along_axis(digitize_1d, axis, x, n_bins)
    else:
        kwargs["n_bins"] = n_bins
        kwargs["encode"] = "ordinal"
        kwargs["subsample"] = None
        return np.apply_along_axis(digitize_sklearn, axis, x, **kwargs)


partial(jax.jit, static_argnums=(1, 2))


def normalize(x, to_min=0.0, to_max=1.0):
    """Normalize the array x between to_min and to_max.

    Parameters
    ----------
    x : array_like
        The array to normalize
    to_min : int/float | 0.
        Minimum of returned array
    to_max : int/float | 1.
        Maximum of returned array

    Returns
    -------
    xn : array_like
        The normalized array
    """
    x_min, x_max = jnp.nanmin(x), jnp.nanmax(x)
    x_n = to_min + (x - x_min) * (to_max - to_min) / (x_max - x_min)
    # x_n = to_max - (((to_max - to_min) * (x_max - x)) / (x_max - x_min))

    return x_n


def get_nbest_mult(
    hoi,
    model=None,
    orders=None,
    multiplets=None,
    n_best=5,
    minsize=None,
    maxsize=None,
    names=None,
):
    """Get the n best multiplets.

    This function requires pandas to be installed.

    Parameters
    ----------
    hoi : array_like
        Array of higher-order information.
    model : hoi.metrics, optional
        Model used to compute the higher-order information. The default is
        None.
    orders: array_like, optional
        Order associated to each multiplet. The default is None.
    multiplets: array_like, optional
        Combination of features. The default is None.
    n_best : int, optional
        Number of best multiplets to return. The default is 5.
    minsize : int, optional
        Minimum size of the multiplets. The default is None.
    maxsize : int, optional
        Maximum size of the multiplets. The default is None.
    names : list, optional
        List of names of the variables. The default is None.

    Returns
    -------
    df_best : pandas.DataFrame
        Dataframe containing the n best multiplets.
    """
    import pandas as pd

    hoi = np.asarray(hoi).squeeze()
    if not hoi.ndim:
        hoi = hoi.reshape(-1)

    # get order and multiplets
    if model:
        orders = model.order
        multiplets = model.multiplets

    # get computed orders
    if minsize is None:
        minsize = orders.min()
    if maxsize is None:
        maxsize = orders.max()

    # get the data at selected order
    indices = np.arange(hoi.shape[0])
    keep_order = np.logical_and(orders >= minsize, orders <= maxsize)

    # merge into a dataframe
    df = pd.DataFrame(
        {
            "hoi": hoi[keep_order],
            "index": indices[keep_order],
            "order": orders[keep_order],
        }
    ).sort_values(by="hoi", ascending=False)

    # df selection
    is_syn = df["hoi"] < 0
    df_syn = df.loc[is_syn]
    df_syn = df_syn.iloc[-min(n_best, len(df_syn)) : :]
    df_red = df.loc[~is_syn]
    df_red = df_red.iloc[: min(n_best, len(df_red))]
    df_best = pd.concat((df_red, df_syn)).reset_index(drop=True)

    # reorder columns
    df_best = df_best[["index", "order", "hoi"]]

    # multiplets selection
    mults = []
    for m in multiplets[df_best["index"].values, :]:
        mults.append(m[m != -1])
    df_best["multiplet"] = mults

    # find names
    if names is not None:
        cols = np.asarray(names)
        names = []
        for c in df_best["multiplet"].values:
            c = np.asarray(c)
            names.append(" / ".join(cols[c].tolist()))
        df_best["names"] = names

    return df_best



from functools import partial

import numpy as np

import jax
import jax.numpy as jnp

from hoi.metrics.base_hoi import HOIEstimator
from hoi.core.entropies import prepare_for_it
from hoi.core.mi import get_mi, compute_mi_comb
from hoi.utils.progressbar import get_pbar


class RedundancyMMI(HOIEstimator):
    """Redundancy estimated using the Minimum Mutual Information.

    Parameters
    ----------
    x : array_like
        Standard NumPy arrays of shape (n_samples, n_features) or
        (n_samples, n_features, n_variables)
    y : array_like
        The feature of shape (n_samples,)
    multiplets : list | None
        List of multiplets to compute. Should be a list of multiplets, for
        example [(0, 1, 2), (2, 7, 8, 9)]. By default, all multiplets are
        going to be computed.
    """

    __name__ = "Redundancy MMI"
    _encoding = True
    _positive = "redundancy"
    _negative = "null"
    _symmetric = False

    def __init__(self, x, y, multiplets=None, verbose=None):
        HOIEstimator.__init__(
            self,
            x=x,
            y=y,
            multiplets=multiplets,
            verbose=verbose,
        )

    def fit(
        self, minsize=2, maxsize=None, method="gc", samples=None, **kwargs
    ):
        """Redundancy Index.

        Parameters
        ----------
        minsize, maxsize : int | 2, None
            Minimum and maximum size of the multiplets
        method : {'gc', 'binning', 'knn', 'kernel', callable}
            Name of the method to compute entropy. Use either :

                * 'gc': gaussian copula entropy [default]. See
                  :func:`hoi.core.entropy_gc`
                * 'gauss': gaussian entropy. See :func:`hoi.core.entropy_gauss`
                * 'binning': binning-based estimator of entropy. Note that to
                  use this estimator, the data have be to discretized. See
                  :func:`hoi.core.entropy_bin`
                * 'knn': k-nearest neighbor estimator. See
                  :func:`hoi.core.entropy_knn`
                * 'kernel': kernel-based estimator of entropy
                  see :func:`hoi.core.entropy_kernel`
                * A custom entropy estimator can be provided. It should be a
                  callable function written with Jax taking a single 2D input
                  of shape (n_features, n_samples) and returning a float.

        samples : np.ndarray
            List of samples to use to compute HOI. If None, all samples are
            going to be used.
        kwargs : dict | {}
            Additional arguments are sent to each MI function

        Returns
        -------
        hoi : array_like
            The NumPy array containing values of higher-order interactions of
            shape (n_multiplets, n_variables)
        """
        # ________________________________ I/O ________________________________
        # check minsize and maxsize
        minsize, maxsize = self._check_minmax(max(minsize, 2), maxsize)

        # prepare the data for computing mi
        x, kwargs = prepare_for_it(self._x, method, samples=samples, **kwargs)
        x, y = self._split_xy(x)

        # prepare mi functions
        mi_fcn = jax.vmap(get_mi(method=method, **kwargs))
        compute_mi = partial(compute_mi_comb, mi=mi_fcn)

        # get multiplet indices and order
        h_idx, order = self.get_combinations(minsize, maxsize=maxsize)

        # get progress bar
        pbar = get_pbar(
            iterable=range(order.min(), order.max() + 1), leave=False
        )

        # _______________________________ HOI _________________________________

        # compute mi I(x_{1}; y), ..., I(x_{n}; y)
        _, i_xiy = jax.lax.scan(
            compute_mi, (x, y), jnp.arange(x.shape[1]).reshape(-1, 1)
        )

        # prepare outputs
        offset = 0
        hoi = jnp.zeros((len(order), self.n_variables), dtype=jnp.float32)
        for msize in pbar:
            pbar.set_description(desc="RedMMI order %s" % msize, refresh=False)

            # combinations of features
            _h_idx = h_idx[order == msize, 0:msize]

            # compute hoi
            _hoi = i_xiy[_h_idx, :].min(1)

            # fill variables
            n_combs = _h_idx.shape[0]
            hoi = hoi.at[offset : offset + n_combs, :].set(_hoi)

            # updates
            offset += n_combs

        return np.asarray(hoi)
